PARSER_TEMPLATE = """from abc import ABC
from pyjapt import ShiftReduceParser
from %s import %s


class %s(ShiftReduceParser, ABC):
    def __init__(self, verbose=False):
        self.G = G
        self.verbose = verbose
        self.action = self.__action_table()
        self.goto = self.__goto_table()
        self._errors = []

    @staticmethod
    def __action_table():
        return %s

    @staticmethod
    def __goto_table():
        return %s
"""

LEXER_TEMPLATE = """import re

from pyjapt import Token, Lexer
from %s import %s


class %s(Lexer):
    def __init__(self):
        self.lineno = 1
        self.column = 1
        self.position = 0
        self.token = Token('', '', 0, 0)
        self.pattern = re.compile(r'%s')
        self.token_rules = %s
        self.error_handler = %s
        self.contain_errors = False
        self.eof = '%s'
    
    def __call__(self, text):
        return %s
"""


class LRParserSerializer:
    @staticmethod
    def build(parser, parser_class_name, grammar_module_name, grammar_variable_name):
        action, goto = LRParserSerializer._build_parsing_tables(parser, grammar_variable_name)
        content = PARSER_TEMPLATE % (grammar_module_name, grammar_variable_name, parser_class_name, action, goto)
        try:
            with open('parsertab.py', 'x') as f:
                f.write(content)
        except FileExistsError:
            with open('parsertab.py', 'w') as f:
                f.write(content)

    @staticmethod
    def _build_parsing_tables(parser, variable_name):
        s1 = '{\n'
        for (state, symbol), (act, tag) in parser.action.items():
            s1 += f'            ({state}, {variable_name}["{symbol}"]): '

            if act == 'SHIFT':
                s1 += f'("{act}", {tag}),\n'
            elif act == 'REDUCE':
                s1 += f'("{act}", {variable_name}["{repr(tag)}"]),\n'
            else:
                s1 += f'("{act}", None),\n'
        s1 += '        }'
        s2 = '{\n'
        for (state, symbol), dest in parser.goto.items():
            s2 += f'            ({state}, {variable_name}["{symbol}"]): {dest},\n'
        s2 += '        }'

        return s1, s2


class LexerSerializer:
    @staticmethod
    def build(lexer, lexer_class_name, grammar_module_name, grammar_variable_name):
        pattern = lexer.pattern

        token_rules = f"{{key: rule for key, (_, _, rule) in {grammar_variable_name}.terminal_rules.items() if rule " \
                      f"is not None}}"

        error_handler = f"{grammar_variable_name}.lexical_error_handler if " \
                        f"{grammar_variable_name}.lexical_error_handler is not None else self.error "

        call_return = f"[Token(t.lex, {grammar_variable_name}[t.token_type], t.line, t.column) for t in " \
                      f"self.tokenize(text)] "

        content = LEXER_TEMPLATE % (
            grammar_module_name, grammar_variable_name, lexer_class_name, pattern, token_rules, error_handler,
            lexer.eof, call_return,
        )

        try:
            with open('lexertab.py', 'x') as f:
                f.write(content)
        except FileExistsError:
            with open('lexertab.py', 'w') as f:
                f.write(content)



from typing import Any, Dict, Set, List, Tuple, Union


class State:
    def __init__(self, state: Any, final: bool = False):
        self.state: Any = state
        self.final: bool = final
        self.transitions: Dict[str, List['State']] = {}
        self.epsilon_transitions: Set['State'] = set()
        self.tag = None

    def has_transition(self, symbol):
        return symbol in self.transitions

    def add_transition(self, symbol: str, state: 'State'):
        try:
            self.transitions[symbol].append(state)
        except KeyError:
            self.transitions[symbol] = [state]
        return self

    def add_epsilon_transition(self, state: 'State'):
        self.epsilon_transitions.add(state)
        return self

    def recognize(self, string: str):
        states = self.epsilon_closure
        for symbol in string:
            states = self.move_by_state(symbol, *states)
            states = self.epsilon_closure_by_state(*states)

        return any(s.final for s in states)

    def to_deterministic(self) -> 'State':
        closure = self.epsilon_closure
        start = State(tuple(closure), any(s.final for s in closure))

        closures = [closure]
        states = [start]
        pending = [start]

        while pending:
            state = pending.pop()
            symbols = {symbol for s in state.state for symbol in s.transitions}

            for symbol in symbols:
                move = self.move_by_state(symbol, *state.state)
                closure = self.epsilon_closure_by_state(*move)

                if closure not in closures:
                    new_state = State(tuple(closure), any(s.final for s in closure))
                    closures.append(closure)
                    states.append(new_state)
                    pending.append(new_state)
                else:
                    index = closures.index(closure)
                    new_state = states[index]

                state.add_transition(symbol, new_state)

        return start

    @staticmethod
    def from_nfa(nfa, get_states: bool = False) -> Union['State', Tuple['State', List['State']]]:
        states = []
        for n in range(nfa.states):
            state = State(n, n in nfa.finals)
            states.append(state)

        for (origin, symbol), destinations in nfa.map.items():
            origin = states[origin]
            origin[symbol] = [states[d] for d in destinations]

        if get_states:
            return states[nfa.start], states
        return states[nfa.start]

    @staticmethod
    def move_by_state(symbol, *states):
        return {s for state in states if state.has_transition(symbol) for s in state[symbol]}

    @staticmethod
    def epsilon_closure_by_state(*states):
        closure = set(states)

        n = 0
        while n != len(closure):
            n = len(closure)
            tmp = [s for s in closure]
            for s in tmp:
                for epsilon_state in s.epsilon_transitions:
                    closure.add(epsilon_state)
        return closure

    @property
    def epsilon_closure(self):
        return self.epsilon_closure_by_state(self)

    def get(self, symbol):
        target = self.transitions[symbol]
        assert len(target) == 1
        return target[0]

    def __getitem__(self, symbol):
        if symbol == '':
            return self.epsilon_transitions
        try:
            return self.transitions[symbol]
        except KeyError:
            return None

    def __setitem__(self, symbol, value):
        if symbol == '':
            self.epsilon_transitions = value
        else:
            self.transitions[symbol] = value

    def __repr__(self):
        return str(self)

    def __str__(self):
        return str(self.state)

    def __hash__(self):
        return hash(self.state)

    def __iter__(self):
        yield from self.__visit()

    def __visit(self, visited=None):
        if visited is None:
            visited = set()
        elif self in visited:
            return

        visited.add(self)
        yield self

        for destinations in self.transitions.values():
            for node in destinations:
                yield from node.__visit(visited)
        for node in self.epsilon_transitions:
            yield from node.__visit(visited)



import json
import re
import sys
from typing import List, FrozenSet, Optional, Tuple, Iterable, Callable, Dict, Set

from pyjapt.automata import State
from pyjapt.lexing import Lexer, Token
from pyjapt.serialization import LRParserSerializer, LexerSerializer
from pyjapt.utils import ContainerSet

TerminalRule = Callable[[Lexer], Optional[Token]]
ProductionRule = Callable[['RuleList'], object]


class GrammarError(Exception):
    def __init__(self, *args):
        super().__init__(args)
        self.text = args[0]


class Symbol:
    def __init__(self, name: str, grammar: Optional['Grammar']):
        self.name: str = name
        self.grammar: 'Grammar' = grammar

    @property
    def is_epsilon(self) -> bool:
        return False

    @property
    def is_terminal(self) -> bool:
        return False

    @property
    def is_non_terminal(self) -> bool:
        return False

    def __str__(self):
        return self.name

    def __repr__(self):
        return repr(self.name)

    def __add__(self, other):
        if isinstance(other, Symbol):
            return Sentence(self, other)

        raise TypeError(other)

    def __or__(self, other):
        if isinstance(other, Sentence):
            return SentenceList(Sentence(self), other)
        raise TypeError(other)

    def __len__(self):
        return 1


class NonTerminal(Symbol):
    def __init__(self, name: str, grammar: 'Grammar'):
        super().__init__(name, grammar)
        self.productions: List['Production'] = []
        self.error_productions: List['Production'] = []

    def __mod__(self, other) -> 'NonTerminal':
        if isinstance(other, str):
            if other:
                p = Production(self, Sentence(*(self.grammar[s] for s in other.split())))
            else:
                p = Production(self, self.grammar.EPSILON)
            self.grammar.add_production(p)
            return self

        if isinstance(other, Symbol):
            p = Production(self, Sentence(other))
            self.grammar.add_production(p)
            return self

        if isinstance(other, Sentence):
            p = Production(self, other)
            self.grammar.add_production(p)
            return self

        if isinstance(other, tuple):
            assert len(other) > 1

            if isinstance(other[0], str):
                if other[0]:
                    other = Sentence(*(self.grammar[s] for s in other[0].split())), other[1]
                else:
                    other = self.grammar.EPSILON, other[1]

            if isinstance(other[0], Symbol) and not isinstance(other[0], Sentence):
                other = Sentence(other[0]), other[1]

            if isinstance(other[0], Sentence):
                p = Production(self, other[0], other[1])
            else:
                raise TypeError("Valid types for a production are 'Symbol', 'Sentence' or 'str'")

            self.grammar.add_production(p)
            return self

        if isinstance(other, SentenceList):
            for s in other:
                p = Production(self, s)
                self.grammar.add_production(p)
            return self

        raise TypeError(other)

    @property
    def is_non_terminal(self) -> bool:
        return True


class Terminal(Symbol):
    def __init__(self, name: str, grammar: 'Grammar'):
        super().__init__(name, grammar)

    @property
    def is_terminal(self) -> bool:
        return True


class PropagationTerminal(Symbol):
    def __init__(self):
        super().__init__('#', None)


class ErrorTerminal(Terminal):
    def __init__(self, grammar: 'Grammar'):
        super().__init__('error', grammar)


class EOF(Terminal):
    def __init__(self, grammar: 'Grammar'):
        super().__init__('$', grammar)


class Sentence:
    def __init__(self, *args):
        self.symbols = tuple(x for x in args if not x.is_epsilon)
        self.hash = hash(self.symbols)

    def __len__(self):
        return len(self.symbols)

    def __add__(self, other) -> 'Sentence':
        if isinstance(other, Symbol):
            return Sentence(*(self.symbols + (other,)))

        if isinstance(other, Sentence):
            return Sentence(*(self.symbols + other.symbols))

        raise TypeError(other)

    def __or__(self, other) -> 'SentenceList':
        if isinstance(other, Sentence):
            return SentenceList(self, other)

        if isinstance(other, Symbol):
            return SentenceList(self, Sentence(other))

        raise TypeError(other)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return ("%s " * len(self.symbols) % tuple(self.symbols)).strip()

    def __iter__(self):
        return iter(self.symbols)

    def __contains__(self, item):
        return item in self.symbols

    def __getitem__(self, index):
        return self.symbols[index]

    def __eq__(self, other) -> bool:
        return self.symbols == other.symbols

    def __hash__(self):
        return self.hash

    @property
    def is_epsilon(self):
        return False


class SentenceList:
    def __init__(self, *args):
        self._sentences = list(args)

    def add(self, symbol):
        if not symbol and (symbol is None or not symbol.is_epsilon):
            raise ValueError(symbol)

        self._sentences.append(symbol)

    def __iter__(self):
        return iter(self._sentences)

    def __or__(self, other):
        if isinstance(other, Sentence):
            self.add(other)
            return self

        if isinstance(other, Symbol):
            return self | Sentence(other)


class Epsilon(Terminal, Sentence):
    def __init__(self, grammar: 'Grammar'):
        super().__init__('epsilon', grammar)

    def __str__(self):
        return "e"

    def __repr__(self):
        return 'epsilon'

    def __iter__(self):
        yield from ()

    def __len__(self):
        return 0

    def __add__(self, other):
        return other

    def __eq__(self, other):
        return isinstance(other, (Epsilon,))

    def __hash__(self):
        return hash("")

    @property
    def is_epsilon(self):
        return True


class Production:
    def __init__(self, non_terminal: NonTerminal, sentence: Sentence,
                 rule: Optional[Callable[['RuleList'], object]] = None):
        self.left: NonTerminal = non_terminal
        self.right: Sentence = sentence
        self.rule: Optional[Callable[['RuleList'], object]] = rule

    @property
    def is_epsilon(self) -> bool:
        return self.right.is_epsilon

    def __str__(self):
        return '%s -> %s' % (self.left, self.right)

    def __repr__(self):
        return '%s -> %s' % (self.left, self.right)

    def __iter__(self):
        yield self.left
        yield self.right

    def __eq__(self, other):
        return isinstance(other, Production) and self.left == other.left and self.right == other.right

    def __hash__(self):
        return hash((self.left, self.right))


class Grammar:
    def __init__(self):
        self.productions: List[Production] = []
        self.non_terminals: List[NonTerminal] = []
        self.terminals: List[Terminal] = []
        self.start_symbol: Optional[NonTerminal] = None

        self.ERROR: ErrorTerminal = ErrorTerminal(self)
        self.EPSILON: Epsilon = Epsilon(self)
        self.EOF: EOF = EOF(self)

        self.lexical_error_handler = None  # type: Optional[Callable[[Lexer], None]]
        self.terminal_rules = {}  # type: Dict[str, Tuple[str, bool, Optional[Callable[[Lexer], Optional[Token]]]]]
        self.production_dict = {}  # type: Dict[str, Production]
        self.symbol_dict = {'$': self.EOF, 'error': self.ERROR}  # type: Dict[str, Symbol]

    def add_terminal(self, name: str, regex: str = None, rule: Optional[TerminalRule] = None) -> Terminal:
        """
        Create a terminal for the grammar with its respective regular expression

        If not regex is given then the regular expression will be the literal name of the Terminal

        If the regex is given then de param name most be a valid python identifier

        All terminals with a not given regex will be marked as `literals` and then it will have a higher priority in
        the lexer recognition above the others, and the reserved set will be sorted by the length of the regular
        expression by descending

        The given rule must be a function

        :param name: str

        :param regex: str

        :param rule: function to handle the arrival of a token during the lexical process

        :return: Terminal
        """
        name = name.strip()
        if not name:
            raise Exception("Empty name")

        assert name not in self.symbol_dict, f'Terminal {name} already defined in grammar'

        literal = False
        if regex is None:
            regex = re.escape(name)
            literal = True

        term = Terminal(name, self)
        self.terminals.append(term)
        self.symbol_dict[name] = term
        self.terminal_rules[name] = regex, literal, rule
        return term

    def add_terminal_error(self):
        self.terminals.append(self.ERROR)

    def add_terminals(self, names: str) -> Tuple[Terminal, ...]:
        return tuple(self.add_terminal(x) for x in names.strip().split())

    def add_non_terminal(self, name: str, start_symbol: bool = False) -> NonTerminal:
        """
        Add and return a new non-terminal, use the start_symbol parameter for define the started symbol in th grammar,
        if two non-terminal are marked as started symbol an AssertionError will be raised

        :param name: str

        :param start_symbol: bool

        :return: NonTerminal created
        """
        name = name.strip()
        if not name:
            raise Exception("Empty name")

        assert name not in self.symbol_dict, f'Non-Terminal {name} already defined in grammar'

        term = NonTerminal(name, self)

        if start_symbol:
            if self.start_symbol is None:
                self.start_symbol = term
            else:
                raise Exception("Cannot define more than one start symbol.")

        self.non_terminals.append(term)
        self.symbol_dict[name] = term
        return term

    def add_non_terminals(self, names: str) -> Tuple[NonTerminal, ...]:
        return tuple(self.add_non_terminal(x) for x in names.strip().split())

    def add_production(self, production: Production):
        production.left.productions.append(production)
        self.productions.append(production)
        self.production_dict[repr(production)] = production

    def production(self, *productions: str) -> Callable[[ProductionRule], ProductionRule]:
        """
        Return a function to decorate a method that will be used for as production rule

        :param productions: is a sequence of strings representing the production to attach the decorated function,
                       the string has the form '<non-terminal name> -> <symbols separated with white spaces>'

        :return: a function to decorate the production
        """

        def decorator(rule: ProductionRule) -> ProductionRule:
            for production in productions:
                head, body = production.split('->')
                head = self[head.strip()]
                head %= body.strip(), rule
            return rule

        return decorator

    def terminal(self, name: str, regex: str) -> Callable[[TerminalRule], TerminalRule]:
        """
        Return a function to decorate a method that will be used for as terminal rule in tokenizer process

        :param name: the name of a terminal

        :param regex: the regex of the terminal

        :return: a function to decorate the production
        """

        def decorator(rule: TerminalRule) -> TerminalRule:
            self.add_terminal(name, regex, rule)
            return rule

        return decorator

    def lexical_error(self, handler: Callable[[Lexer], None]):
        self.lexical_error_handler = handler
        return handler

    def augmented_grammar(self, force: bool = False):
        if not self.is_augmented_grammar or force:
            grammar = self.copy()
            start_symbol = grammar.start_symbol
            grammar.start_symbol = None
            new_start_symbol = grammar.add_non_terminal('S\'', True)
            new_start_symbol %= start_symbol + grammar.EPSILON, lambda x: x
            return grammar
        else:
            return self.copy()

    def copy(self):
        grammar = Grammar()
        grammar.productions = self.productions.copy()
        grammar.non_terminals = self.non_terminals.copy()
        grammar.terminals = self.terminals.copy()
        grammar.start_symbol = self.start_symbol
        grammar.EPSILON = self.EPSILON
        grammar.ERROR = self.ERROR
        grammar.EOF = self.EOF
        grammar.symbol_dict = self.symbol_dict.copy()

        return grammar

    def get_lexer(self) -> Lexer:
        items = self.terminal_rules.items()
        f1 = filter(lambda x: x[1][2] is not None, items)
        f2 = filter(lambda x: x[1][2] is None and not x[1][1], items)
        f3 = filter(lambda x: x[1][2] is None and x[1][1], items)

        ruled_tokens = list(f1)
        not_literal_tokens = sorted(f2, key=lambda x: len(x[1][0]), reverse=True)
        literal_tokens = sorted(f3, key=lambda x: len(x[1][0]), reverse=True)

        table = (
                [(name, regex) for name, (regex, _, _) in ruled_tokens] +
                [(name, regex) for name, (regex, _, _) in not_literal_tokens] +
                [(None, regex) for _, (regex, _, _) in literal_tokens]
        )

        return Lexer(table, self.EOF.name,
                     {s: r for s, (_, _, r) in items if r is not None}, self.lexical_error_handler)

    def get_parser(self, name: str, verbose: bool = False):
        if name == 'slr':
            return SLRParser(self, verbose)

        if name == 'lalr1':
            return LALR1Parser(self, verbose)

        if name == 'lr1':
            return LR1Parser(self, verbose)

        raise ValueError(name)  # TODO: create a custom error

    def serialize_lexer(self, class_name: str, grammar_module_name: str, grammar_variable_name: str = 'G'):
        LexerSerializer.build(self.get_lexer(), class_name, grammar_module_name, grammar_variable_name)

    def serialize_parser(self, parser_type: str, class_name: str, grammar_module_name: str,
                         grammar_variable_name: str = 'G'):
        LRParserSerializer.build(self.get_parser(parser_type), class_name, grammar_module_name, grammar_variable_name)

    @property
    def is_augmented_grammar(self):
        augmented = 0
        for left, _ in self.productions:
            if self.start_symbol == left:
                augmented += 1
        if augmented <= 1:
            return True
        else:
            return False

    def to_json(self):

        productions = []

        for p in self.productions:
            head = p.left.name

            body = []

            for s in p.right:
                body.append(s.Name)

            productions.append({'Head': head, 'Body': body})

        d = {'NonTerminals': [symbol.name for symbol in self.non_terminals],
             'Terminals': [symbol.name for symbol in self.terminals],
             'Productions': productions}

        # [{'Head':p.Left.Name, "Body": [s.Name for s in p.Right]} for p in self.Productions]
        return json.dumps(d)

    @staticmethod
    def from_json(data):
        data = json.loads(data)

        grammar = Grammar()
        dic = {'epsilon': grammar.EPSILON}

        for term in data['Terminals']:
            dic[term] = grammar.add_terminal(term)

        for noTerm in data['NonTerminals']:
            dic[noTerm] = grammar.add_non_terminal(noTerm)

        for p in data['Productions']:
            head = p['Head']
            dic[head] %= Sentence(*[dic[term] for term in p['Body']])

        return grammar

    def __getitem__(self, item):
        try:
            try:
                return self.symbol_dict[item]
            except KeyError:
                return self.production_dict[item]
        except KeyError:
            return None

    def __str__(self):
        mul = '%s, '
        ans = 'Non-Terminals:\n\t'
        non_terminals = mul * (len(self.non_terminals) - 1) + '%s\n'
        ans += non_terminals % tuple(self.non_terminals)
        ans += 'Terminals:\n\t'
        terminals = mul * (len(self.terminals) - 1) + '%s\n'
        ans += terminals % tuple(self.terminals)
        ans += 'Productions:\n\t'
        ans += str(self.productions)
        return ans


class Item:
    def __init__(self, production: Production, pos: int, lookaheads: Iterable[Symbol] = None):
        if lookaheads is None:
            lookaheads = []
        self.production: Production = production
        self.pos: int = pos
        self.lookaheads: FrozenSet[Symbol] = frozenset(lookaheads)

    def __str__(self):
        s = str(self.production.left) + " -> "
        if len(self.production.right) > 0:
            for i, _ in enumerate(self.production.right):
                if i == self.pos:
                    s += "."
                s += str(self.production.right[i]) + " "
            if self.pos == len(self.production.right):
                s += "."
        else:
            s += "."
        s += ", " + str(self.lookaheads)[10:-1]
        return s

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
                (self.pos == other.pos) and
                (self.production == other.production) and
                (set(self.lookaheads) == set(other.lookaheads))
        )

    def __hash__(self):
        return hash((self.production, self.pos, self.lookaheads))

    @property
    def is_reduce_item(self) -> bool:
        return len(self.production.right) == self.pos

    @property
    def next_symbol(self) -> Optional[Symbol]:
        if self.pos < len(self.production.right):
            return self.production.right[self.pos]
        else:
            return None

    def next_item(self) -> Optional['Item']:
        if self.pos < len(self.production.right):
            return Item(self.production, self.pos + 1, self.lookaheads)
        else:
            return None

    def preview(self, skip: int = 1) -> List[Symbol]:
        unseen = self.production.right[self.pos + skip:]
        return [unseen + (lookahead,) for lookahead in self.lookaheads]

    def center(self) -> 'Item':
        return Item(self.production, self.pos)


class RuleList:
    def __init__(self, parser, rules):
        self.__parser = parser
        self.__list = rules

    def __iter__(self):
        return iter(self.__list)

    def __getitem__(self, item):
        return self.__list[item]

    def __setitem__(self, key, value):
        self.__list[key] = value

    def __len__(self):
        return len(self.__list)

    def set_error(self, index: int, message: str):
        self.__parser.set_error(self[index].line, self[index].column, message)


##############################
# Compute Firsts and Follows #
##############################
def compute_local_first(firsts, alpha):
    first_alpha = ContainerSet()

    try:
        alpha_is_epsilon = alpha.is_epsilon
    except AttributeError:
        alpha_is_epsilon = False

    if alpha_is_epsilon:
        first_alpha.set_epsilon()
    else:
        for symbol in alpha:
            first_symbol = firsts[symbol]
            first_alpha.update(first_symbol)
            if not first_symbol.contains_epsilon:
                break
        else:
            first_alpha.set_epsilon()

    return first_alpha


def compute_firsts(grammar: Grammar):
    firsts = {}
    change = True

    for terminal in grammar.terminals:
        firsts[terminal] = ContainerSet(terminal)

    for non_terminal in grammar.non_terminals:
        firsts[non_terminal] = ContainerSet()

    while change:
        change = False

        # P: X -> alpha
        for production in grammar.productions:
            x, alpha = production

            first_x = firsts[x]

            try:
                first_alpha = firsts[alpha]
            except KeyError:
                first_alpha = firsts[alpha] = ContainerSet()

            local_first = compute_local_first(firsts, alpha)

            change |= first_alpha.hard_update(local_first)
            change |= first_x.hard_update(local_first)

    return firsts


def compute_follows(grammar: Grammar, firsts):
    follows = {}
    change = True

    local_firsts = {}

    # init Follow(Vn)
    for non_terminal in grammar.non_terminals:
        follows[non_terminal] = ContainerSet()
    follows[grammar.start_symbol] = ContainerSet(grammar.EOF)

    while change:
        change = False

        # P: X -> alpha
        for production in grammar.productions:
            x, alpha = production

            follow_x = follows[x]

            for i, symbol_y in enumerate(alpha):
                # X -> zeta Y beta
                if symbol_y.is_non_terminal:
                    follow_y = follows[symbol_y]
                    try:
                        first_beta = local_firsts[alpha, i]
                    except KeyError:
                        first_beta = local_firsts[alpha, i] = compute_local_first(firsts, alpha[i + 1:])
                    # First(beta) - { epsilon } subset of Follow(Y)
                    change |= follow_y.update(first_beta)
                    # beta ->* epsilon or X -> zeta Y ? Follow(X) subset of Follow(Y)
                    if first_beta.contains_epsilon:
                        change |= follow_y.update(follow_x)
    # Follow(Vn)
    return follows


#########################
# LR0 AUTOMATA BUILDING #
#########################
def closure_lr0(items: Iterable[Item]):
    closure = set(items)

    pending = set(items)
    while pending:
        current = pending.pop()
        symbol = current.next_symbol

        if current.is_reduce_item or symbol.is_terminal:
            continue

        new_items = set(Item(p, 0) for p in symbol.productions)  # if Item(p, 0) not in closure]
        pending |= new_items - closure
        closure |= new_items
    return frozenset(closure)


def goto_lr0(items: Iterable[Item], symbol: Symbol) -> FrozenSet[Item]:
    return frozenset(item.next_item() for item in items if item.next_symbol == symbol)


def build_lr0_automaton(grammar: Grammar, just_kernel: bool = False) -> State:
    assert len(grammar.start_symbol.productions) == 1, 'Grammar must be augmented'

    start_production = grammar.start_symbol.productions[0]
    start_item = Item(start_production, 0)
    start = frozenset([start_item])

    if not just_kernel:
        automaton = State(closure_lr0(start), True)
    else:
        automaton = State(start, True)

    pending = [start]
    visited = {start: automaton}

    symbols = grammar.terminals + grammar.non_terminals
    while pending:
        current = pending.pop()
        current_state = visited[current]
        current_closure = current_state.state if not just_kernel else closure_lr0(current)
        for symbol in symbols:
            kernel = goto_lr0(current_closure, symbol)

            if kernel == frozenset():
                continue

            try:
                next_state = visited[kernel]
            except KeyError:
                next_state = visited[kernel] = State(closure_lr0(kernel), True) if not just_kernel else State(kernel,
                                                                                                              True)
                pending.append(kernel)

            current_state.add_transition(symbol.name, next_state)
    return automaton


#########################
# LR1 AUTOMATA BUILDING #
#########################
def compress(items: Iterable[Item]) -> Set[Item]:
    centers = {}

    for item in items:
        center = item.center()
        try:
            lookaheads = centers[center]
        except KeyError:
            centers[center] = lookaheads = set()
        lookaheads.update(item.lookaheads)

    return set(Item(x.production, x.pos, set(lookaheads)) for x, lookaheads in centers.items())


def expand(item: Item, firsts):
    next_symbol = item.next_symbol
    if next_symbol is None or not next_symbol.is_non_terminal:
        return []

    lookaheads = ContainerSet()

    for preview in item.preview():
        local_first = compute_local_first(firsts, preview)
        lookaheads.update(local_first)

    assert not lookaheads.contains_epsilon

    return [Item(p, 0, lookaheads) for p in next_symbol.productions]


def closure_lr1(items, firsts):
    closure = set(items)
    pending = set(items)
    while pending:
        new_items = set(expand(pending.pop(), firsts))
        pending |= new_items - closure
        closure |= new_items
    return compress(closure)


def goto_lr1(items, symbol, firsts=None, just_kernel=False):
    assert just_kernel or firsts is not None, '`firsts` must be provided if `just_kernel=False`'
    items = frozenset(item.next_item() for item in items if item.next_symbol == symbol)
    return items if just_kernel else closure_lr1(items, firsts)


def build_lr1_automaton(grammar, firsts=None):
    assert len(grammar.start_symbol.productions) == 1, 'Grammar must be augmented'

    if not firsts:
        firsts = compute_firsts(grammar)
    firsts[grammar.EOF] = ContainerSet(grammar.EOF)

    start_production = grammar.start_symbol.productions[0]
    start_item = Item(start_production, 0, lookaheads=(grammar.EOF,))
    start = frozenset([start_item])

    closure = closure_lr1(start, firsts)
    automaton = State(frozenset(closure), True)

    pending = [start]
    visited = {start: automaton}

    symbols = grammar.terminals + grammar.non_terminals
    while pending:
        current = pending.pop()
        current_state = visited[current]

        current_closure = current_state.state
        for symbol in symbols:
            kernel = goto_lr1(current_closure, symbol, just_kernel=True)

            if kernel == frozenset():
                continue

            try:
                next_state = visited[kernel]
            except KeyError:
                goto = closure_lr1(kernel, firsts)
                visited[kernel] = next_state = State(frozenset(goto), True)
                pending.append(kernel)
            current_state.add_transition(symbol.name, next_state)

    return automaton


###########################
# LALR1 AUTOMATA BUILDING #
###########################
def determining_lookaheads(state, propagate, table, firsts):
    for item_kernel in state.state:
        closure = closure_lr1([Item(item_kernel.production, item_kernel.pos, ('#',))], firsts)
        for item in closure:
            if item.is_reduce_item:
                continue

            next_state = state.get(item.next_symbol.name)
            next_item = item.next_item().center()
            if '#' in item.lookaheads:
                propagate[state, item_kernel].append((next_state, next_item))
            table[next_state, next_item].extend(item.lookaheads - {'#'})


def build_lalr1_automaton(grammar, firsts=None):
    automaton = build_lr0_automaton(grammar, just_kernel=True)

    if not firsts:
        firsts = compute_firsts(grammar)
    firsts['#'] = ContainerSet('#')
    firsts[grammar.EOF] = ContainerSet(grammar.EOF)

    table = {(state, item): ContainerSet() for state in automaton for item in state.state}
    propagate = {(state, item): [] for state in automaton for item in state.state}

    for state in automaton:
        determining_lookaheads(state, propagate, table, firsts)
    del firsts['#']

    start_item = list(automaton.state).pop()
    table[automaton, start_item] = ContainerSet(grammar.EOF)

    change = True
    while change:
        change = False
        for from_state, from_item in propagate:
            for to_state, to_item in propagate[from_state, from_item]:
                change |= table[to_state, to_item].extend(table[from_state, from_item])

    for state in automaton:
        for item in state.state:
            item.lookaheads = frozenset(table[state, item])

    for state in automaton:
        state.state = frozenset(closure_lr1(state.state, firsts))

    return automaton


#############################
# SLR & LR1 & LALR1 Parsers #
#############################
class ShiftReduceParser:
    SHIFT = 'SHIFT'
    REDUCE = 'REDUCE'
    OK = 'OK'
    contains_errors = False

    def __init__(self, grammar: Grammar, verbose: bool = False):
        self.grammar = grammar
        self.augmented_grammar = grammar.augmented_grammar(True)
        self.firsts = compute_firsts(self.augmented_grammar)
        self.follows = compute_follows(self.augmented_grammar, self.firsts)
        self.automaton = self._build_automaton()
        self.conflicts = []
        self.verbose = verbose

        self.action = {}
        self.goto = {}
        self.shift_reduce_count = 0
        self.reduce_reduce_count = 0
        self._errors = []
        self._build_parsing_table()

        if self.conflicts:
            sys.stderr.write(f"Warning: {self.shift_reduce_count} Shift-Reduce Conflicts\n")
            sys.stderr.write(f"Warning: {self.reduce_reduce_count} Reduce-Reduce Conflicts\n")

    ##############
    # Errors API #
    ##############
    @property
    def errors(self, clean: bool = True):
        return [(m if clean else (r, c, m)) for r, c, m in sorted(self._errors)]

    def set_error(self, line, column, message):
        self._errors.append((line, column, message))

    #############
    #    End    #
    #############

    def _build_parsing_table(self):
        grammar = self.augmented_grammar
        automaton = self.automaton

        for i, node in enumerate(automaton):
            node.id = i

        for node in automaton:
            for item in node.state:
                if item.is_reduce_item:
                    if item.production.left == grammar.start_symbol:
                        self._register(self.action, (node.id, grammar.EOF), (self.OK, None))
                    else:
                        for lookahead in self._lookaheads(item):
                            self._register(self.action, (node.id, lookahead), (self.REDUCE, item.production))
                else:
                    symbol = item.next_symbol
                    if symbol.is_terminal:
                        self._register(self.action, (node.id, symbol), (self.SHIFT, node.get(symbol.name).id))
                    else:
                        self._register(self.goto, (node.id, symbol), node.get(symbol.name).id)

    def _register(self, table, key, value):
        if key in table and table[key] != value:
            action, tag = table[key]
            if action != value[0]:
                if action == self.SHIFT:
                    table[key] = value  # By default shifting if exists a Shift-Reduce Conflict
                self.shift_reduce_count += 1
                self.conflicts.append(('SR', value[1], tag))
            else:
                self.reduce_reduce_count += 1
                self.conflicts.append(('RR', value[1], tag))
        else:
            table[key] = value

    def _build_automaton(self):
        raise NotImplementedError()

    def _lookaheads(self, item):
        raise NotImplementedError()

    def __call__(self, tokens: List[Token]):
        """
        Parse the given TokenList

        :param tokens: List[Token]
        :return: Any
        """
        inserted_error = False
        stack: list = [0]  # The order in stack is [init state] + [symbol, rule, state, ...]
        cursor = 0

        while True:
            if cursor >= len(tokens):
                return

            state = stack[-1]
            lookahead = tokens[cursor]

            if isinstance(lookahead.token_type, str):
                # making the token_type always a terminal
                lookahead.token_type = self.grammar[lookahead.token_type]

            if self.verbose:
                prev = ' '.join([s.name for s in stack if isinstance(s, Symbol)])
                post = ' '.join([tokens[i].lex for i in range(cursor, len(tokens))])
                print(f'{prev} <-> {post}')
                print()

            ##########################
            # Error Handling Section #
            ##########################
            if (state, lookahead.token_type) not in self.action:
                self.contains_errors = True

                if (state, self.grammar.ERROR) in self.action:
                    if self.verbose:
                        print(f'Inserted error token {lookahead,}')

                    inserted_error = True
                    lookahead = Token(lookahead.lex, self.grammar.ERROR, lookahead.line, lookahead.column)
                else:
                    # If an error insertion fails then the parsing process enter into a panic mode recovery
                    sys.stderr.write(
                        f'{lookahead.line, lookahead.column} - SyntacticError: ERROR at or near "{lookahead.lex}"\n')

                    while (state, lookahead.token_type) not in self.action:
                        cursor += 1
                        if cursor >= len(tokens):
                            return
                        lookahead = tokens[cursor]

                    continue
            #######
            # End #
            #######

            action, tag = self.action[state, lookahead.token_type]

            if action == self.SHIFT:
                # in this case tag is an integer
                if self.verbose:
                    print(f'Shift: {lookahead.lex, tag}')

                if not inserted_error:
                    # the rule of an error token is it lexeme
                    stack += [lookahead.token_type, lookahead.lex, tag]
                    cursor += 1
                else:
                    # the rule of an error token is the self token
                    stack += [lookahead.token_type, lookahead, tag]
            elif action == self.REDUCE:
                # in this case tag is a Production
                if self.verbose:
                    print(f'Reduce: {repr(tag)}')

                head, body = tag

                rules = RuleList(self, [None] * (len(body) + 1))
                for i, s in enumerate(reversed(body), 1):
                    state, rules[-i], symbol = stack.pop(), stack.pop(), stack.pop()
                    assert s == symbol, f'ReduceError: in production "{repr(tag)}". Expected {s} instead of {s}'

                if tag.rule is not None:
                    rules[0] = tag.rule(rules)

                state = stack[-1]
                goto = self.goto[state, head]
                stack += [head, rules[0], goto]
            elif action == self.OK:
                return stack[2]
            else:
                raise Exception(f'ParsingError: invalid action {action}')

            inserted_error = False


class SLRParser(ShiftReduceParser):
    def _build_automaton(self):
        return build_lr0_automaton(self.augmented_grammar)

    def _lookaheads(self, item):
        return self.follows[item.production.left]


class LR1Parser(ShiftReduceParser):
    def _build_automaton(self):
        return build_lr1_automaton(self.augmented_grammar, firsts=self.firsts)

    def _lookaheads(self, item):
        return item.lookaheads


class LALR1Parser(LR1Parser):
    def _build_automaton(self):
        return build_lalr1_automaton(self.augmented_grammar, firsts=self.firsts)



class ContainerSet:
    """
    Special class used for handling the existence of epsilon in a set of Symbols. It has all properties of a set
    """

    def __init__(self, *values, contains_epsilon: bool = False):
        self.set: set = set(values)
        self.contains_epsilon: bool = contains_epsilon

    def add(self, value) -> bool:
        """
        Add a new value to the ContainerSet and return true if size was changed
        :param value: value to be added
        :return: True if the item was added
        """
        n = len(self.set)
        self.set.add(value)
        return n != len(self.set)

    def extend(self, values) -> bool:
        n = len(self.set)
        self.set.update(values)
        return n != len(self.set)

    def set_epsilon(self, value=True) -> bool:
        last = self.contains_epsilon
        self.contains_epsilon = value
        return last != self.contains_epsilon

    def update(self, other) -> bool:
        n = len(self.set)
        self.set.update(other.set)
        return n != len(self.set)

    def epsilon_update(self, other) -> bool:
        return self.set_epsilon(self.contains_epsilon | other.contains_epsilon)

    def hard_update(self, other) -> bool:
        return self.update(other) | self.epsilon_update(other)

    def __contains__(self, item):
        return item in self.set

    def __len__(self):
        return len(self.set) + int(self.contains_epsilon)

    def __str__(self):
        return '%s-%s' % (str(self.set), self.contains_epsilon)

    def __repr__(self):
        return str(self)

    def __iter__(self):
        return iter(self.set)

    def __nonzero__(self):
        return len(self) > 0

    def __eq__(self, other):
        if isinstance(other, set):
            return self.set == other
        return (isinstance(other, ContainerSet) and
                self.set == other.set and
                self.contains_epsilon == other.contains_epsilon)



import re
from typing import List, Any, Generator, Tuple, Pattern, Optional, Callable, Dict


class Token:
    """
    A Token class.

    Parameters
    ----------
    lex: str
        Token's lexeme.
    token_type: Enum
        Token's type.
    """

    def __init__(self, lex, token_type, line=0, column=0):
        """
        :param lex: str
        :param token_type: Enum
        :param line: int
        :param column: int
        """
        self.lex: str = lex
        self.token_type: Any = token_type
        self.line: int = line
        self.column: int = column

    def __str__(self):
        return f'{self.token_type}: {self.lex}'

    def __repr__(self):
        return str(self)

    @property
    def is_valid(self):
        return True


class Lexer:
    def __init__(self, table: List[Tuple[str, str]], eof: str,
                 token_rules: Optional[Dict[str, Callable[['Lexer'], Optional[Token]]]] = None,
                 error_handler: Optional[Callable[['Lexer'], None]] = None):
        if token_rules is None:
            token_rules = {}

        if error_handler is None:
            error_handler = self.error

        self.lineno: int = 1  # Current line number
        self.column: int = 1  # Current column in the line
        self.position: int = 0  # Current position in recognition
        self.token: Token = Token('', '', 0, 0)  # Current token in recognition
        self.pattern: Pattern = self._build_regex(table)
        self.token_rules = token_rules  # type: Dict[str, Callable[['Lexer'], Optional[Token]]]
        self.contain_errors: bool = False
        self.error_handler = error_handler if error_handler is not None else self.error
        self.eof: str = eof
        self.errors: List[Tuple[int, int, str]] = []

    def tokenize(self, text: str) -> Generator[Token, None, None]:
        while self.position < len(text):
            match = self.pattern.match(text, pos=self.position)

            if match is None:
                self.contain_errors = True
                self.token = Token(text[self.position], None, self.lineno, self.column)
                self.error_handler(self)
                continue

            lexeme = match.group()
            token_type = match.lastgroup if match.lastgroup is not None else match.group()
            self.token = Token(lexeme, token_type, self.lineno, self.column)

            if token_type in self.token_rules:
                token = self.token_rules[token_type](self)
                if token is not None and isinstance(token, Token):
                    yield token
                continue

            yield self.token

            self.position = match.end()
            self.column += len(match.group())
        yield Token('$', self.eof, self.lineno, self.column)

    def set_error(self, line: int, col: int, error_msg: str):
        self.errors.append((line, col, error_msg))

    @staticmethod
    def error(lexer: 'Lexer') -> None:
        lexer.set_error(lexer.token.line, lexer.token.column, f'LexerError: Unexpected symbol "{lexer.token.lex}"')
        lexer.position += len(lexer.token.lex)
        lexer.column += len(lexer.token.lex)

    @staticmethod
    def _build_regex(table: List[Tuple[str, str]]) -> Pattern:
        return re.compile('|'.join(
            [('(?P<%s>%s)' % (name, regex) if name is not None else '(%s)' % regex) for name, regex in table]))

    def __call__(self, text: str) -> List[Token]:
        return list(self.tokenize(text))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyjapt']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pyjapt',
    'version': '0.1.3',
    'description': 'Just Another Parsing Tool Writ ten in Python',
    'long_description': '# Lexer and LR parser generator  "PyJapt"\n\nPyJapt is a lexer and parser generator developed to provide a solution not only to the creation of these pieces of the compilation process, but also to allow a custom syntactic and lexicographic error handling interface. For its construction we have been inspired by other parser generators such as yacc, bison, ply and antlr, for example.\n\nPyJapt revolves around the concept of grammar.\n\nTo define the nonterminals of the grammar we use the `add_non_terminal ()` method of the `Grammar` class.\n\n```python\nfrom pyjapt import Grammar\n\ng = Grammar()\n\nexpr = g.add_non_terminal(\'expr\', start_symbol=True)\nterm = g.add_non_terminal(\'term\')\nfact = g.add_non_terminal(\'fact\')\n```\n\nTo define the terminals of our grammar we will use the `add_terminal ()` method of the `Grammar` class. This method receives the name of the non-terminal as the first parameter and a regular expression for the lexicographic analyzer as an optional parameter. In case the second parameter is not provided, the regular expression will be the literal name of the terminal.\n\n```python\nplus = g.add_terminals(\'+\')\nminus = g.add_terminals(\'-\')\nstar = g.add_terminals(\'*\')\ndiv = g.add_terminals(\'/\')\n\nnum = g.add_terminal(\'int\', regex=r\'\\d+\')\n```\n\nIf we have a set of terminals whose regular expression matches their own name we can encapsulate them with the `add_terminals()` function of the `Grammar` class.\n\n```python\nplus, minus, star, div = g.add_terminals(\'+ - * /\')\nnum = g.add_terminal(\'int\', regex=r\'\\d+\')\n```\n\nIt may also be the case that we want to apply a rule when a specific terminal is found, for this PyJapt gives us the `terminal()` function decorator of the `Grammar` class that receives the terminal name and regular expression. The decorated function must receive as a parameter a reference to the lexer to be able to modify parameters such as the row and column of the terminals or the parser reading position and return a `Token`, if this token is not returned it will be ignored.\n\n```python\n@g.terminal(\'int\', r\'\\d+\')\ndef id_terminal(lexer):\n    lexer.column += len(lexer.token.lex)\n    lexer.position += len(lexer.token.lex)\n    lexer.token.lex = int(lexer.token.lex)\n    return lexer.token\n```\n\nWe can also use this form of terminal definition to skip certain characters or tokens, we just need to ignore the return in the method.\n\n```python\n##################\n# Ignored Tokens #\n##################\n@g.terminal(\'newline\', r\'\\n+\')\ndef newline(lexer):\n    lexer.lineno += len(lexer.token.lex)\n    lexer.position += len(lexer.token.lex)\n    lexer.column = 1\n\n\n@g.terminal(\'whitespace\', r\' +\')\ndef whitespace(lexer):\n    lexer.column += len(lexer.token.lex)\n    lexer.position += len(lexer.token.lex)\n\n\n@g.terminal(\'tabulation\', r\'\\t+\')\ndef tab(lexer):\n    lexer.column += 4 * len(lexer.token.lex)\n    lexer.position += len(lexer.token.lex)\n```\n\nTo define the productions of our grammar we can use an attributed or not attributed form:\n\n```python\n# This is an unattributed grammar using previously declared variables\nexpr %= expr + plus + term\nexpr %= expr + minus + term\nexpr %= term\n\nterm %= term + star + fact\nterm %= term + div + fact\nterm %= fact\n\nfact %= num\n\n# A little easier to read...\n# Each symbol in the production string must be separated by a blank space\nexpr %= \'expr + term\'\nexpr %= \'expr - term\'\nexpr %= \'term\'\n\nterm %= \'term * factor\'\nterm %= \'term / factor\'\nterm %= \'fact\'\n\nfact %= \'int\'\n\n# This is an attributed grammar\nexpr %= \'expr + term\', lambda s: s[1] + s[3]\nexpr %= \'expr - term\', lambda s: s[1] + s[3]\nexpr %= \'term\', lambda s: s[1]\n\nterm %= \'term * factor\', lambda s: s[1] + s[3]\nterm %= \'term / factor\', lambda s: s[1] + s[3]\nterm %= \'fact\', lambda s: s[1]\n\nfact %= \'int\', lambda s: int(s[1])\n\n# We can also attribute a function to define a semantic rule\n# This function should receive as parameter `s` which is a reference to a\n# special list with the semantic rules of each symbol of the production.\n# To separate the symbol from the head of the body of the expression\n# use the second symbol `->`\n@g.production(\'expr -> expr + term\')\ndef expr_prod(s):\n    print(\'Adding an expression and a term ;)\')\n    return s[1] + s[3]\n\n# We can also assign a rule to many productions with this decorator\n# Please ignore the fact this grammar is ambiguous and pyjapt doesn\'t have support for it ... yet ;)\n@g.production(\'expr -> expr + expr\', \n              \'expr -> expr - expr\', \n              \'expr -> expr * expr\', \n              \'expr -> expr / expr\',\n              \'expr -> ( expr )\' \n              \'expr -> int\')\ndef expr_prod(s):\n    if len(s) != 2:\n        if s[2] == \'+\':\n            return s[1] + s[2]\n    \n        if s[2] == \'-\':\n            return s[1] + s[2]\n        \n        if s[2] == \'*\':\n            return s[1] + s[2]\n        \n        if s[2] == \'/\':\n            return s[1] + s[2]\n    \n        return s[2]\n    return int(s[1])\n```\n\nTo generate the lexer and the parser of the grammar we will use the get_lexer and get_parser methods respectively. In the case of get_parser it receives a string as an argument to find out what type of parser to use. Valid strings are `\'slr\'`,`\' lr1\'`, and `\'lalr1\'`.\n\n```python\ng.get_lexer()\ng.get_parser(\'slr\')\n```\n\nFinally an example of the entire pipeline in the way that we consider the most readable and comfortable to describe a grammar for the language of arithmetic expressions\n\n```python\nfrom pyjapt import Grammar\n\ng = Grammar()\nexpr = g.add_non_terminal(\'expr\', True)\nterm, fact = g.add_non_terminals(\'term fact\')\ng.add_terminals(\'+ - / * ( )\')\ng.add_terminal(\'int\', regex=r\'\\d+\')\n\n@g.terminal(\'whitespace\', r\' +\')\ndef whitespace(_lexer):\n    _lexer.column += len(_lexer.token.lex)\n    _lexer.position += len(_lexer.token.lex)\n\nexpr %= \'expr + term\', lambda s: s[1] + s[3]\nexpr %= \'expr - term\', lambda s: s[1] - s[3]\nexpr %= \'term\', lambda s: s[1]\n\nterm %= \'term * fact\', lambda s: s[1] * s[3]\nterm %= \'term / fact\', lambda s: s[1] // s[3]\nterm %= \'fact\', lambda s: s[1]\n\nfact %= \'( expr )\', lambda s: s[2]\nfact %= \'int\', lambda s: int(s[1])\n\nlexer = g.get_lexer()\n\nparser = g.get_parser(\'slr\')\n\nprint(parser(lexer(\'(2 + 2) * 2 + 2\'))) # prints 10\n```\n## Serialization\n\nWhen the grammar is large enough the parser construction process can be a bottleneck. To solve this problem PyJapt offers a solution by serializing the parser and lexer generated from the grammar into two files `parsetab.py` and` lexertab.py`. These files must refer to the instance of the original grammar.\n\nTo serialize, the name of the variable that contains the instance of Grammar is required, the name of the module where the grammar was written, and the name of the class that the Lexer and Parser will have serialized.\n\n```python\nimport inspect\nfrom pyjapt import Grammar\n\ng = Grammar()\n\n# ...\n\nif __name__ == \'__main__\':\n    module_name = inspect.getmodulename(__file__)\n    g.serialize_lexer(class_name=\'MyLexer\', grammar_module_name=module_name, grammar_variable_name=\'g\')\n    g.serialize_parser(parser_type=\'slr\', class_name=\'MyParser\', grammar_module_name=module_name, grammar_variable_name=\'g\') \n```\n\n## Lexer Construction\n\nAlthough it seems simple, there are some things to keep in mind when defining the terminals of our grammar. When detecting a token, the lexer will first recognize those that have been marked with the `terminal` decorator of the` Grammar` class ( or have a rule assigned ) according to their order of appearance, then those whose regular expression does not match their identifier will follow, and finally the literal terminals (those that do match their identifier with their regular expression). In the last two cases, the order of both will be given by the size of their regular expression ( largest first ).\n\nIf your language has keywords and identifiers a great idea to avoid the rule of the largest regular expression is set a rule to our identifier terminal\n\n```python\nfrom pyjapt import Grammar\n\n\ng = Grammar()\n\nkeywords = g.add_terminals(\n    \'class inherits if then else fi while loop pool let in case of esac new isvoid true false not\')\nkeywords_names = {x.name for x in keywords}\n\n@g.terminal(\'id\', r\'[a-zA-Z_][a-zA-Z0-9_]*\')\ndef id_terminal(lexer):\n    lexer.column += len(lexer.token.lex)\n    lexer.position += len(lexer.token.lex)\n    if lexer.token.lex in keywords_names:\n        # modify the token type ;) \n        lexer.token.token_type = lexer.token.lex\n    return lexer.token\n```\n\n## Handling of lexicographic and syntactic errors\n\nAn important part of the parsing process is handling errors. For this we can do the parser by hand and insert the error report, since techniques such as `Panic Recovery Mode` which implements `PyJapt` only allow the execution of our parser not to stop, to give specific error reports `PyJapt` offers the creation of erroneous productions to report common errors in a programming language such as the lack of a `;`, an unknown operator, etc. For this our grammar must activate the terminal error flag.\n\n```python\ng.add_terminal_error() # Add the error terminal to the grammar.\n\n# Example of a possible error production\n@g.production("instruction -> let id = expr error")\ndef attribute_error(s):\n    # With this line we report the error message\n    # As the semantic rule of s[5] is the token itself (because it is a terminal error), so we have access\n    # to your their, token type, line and column.\n    s.set_error(5, f"{s[5].line, s[5].column} - SyntacticError: Expected \';\' instead of \'{s[5].lex}\'")\n\n    # With this line we allow to continue creating a node of the ast to\n    # be able to detect semantic errors despite syntactic errors\n    return LetInstruction(s[2], s[4])\n```\n\nTo report lexicographical errors the procedure is quite similar we only define a token that contains an error, in this example a multi-line comment that contains an end of string.\n\n```python\n@g.terminal(\'comment_error\', r\'\\(\\*(.|\\n)*$\')\ndef comment_eof_error(lexer):\n    lexer.contain_errors = True\n    lex = lexer.token.lex\n    for s in lex:\n        if s == \'\\n\':\n            lexer.lineno += 1\n            lexer.column = 0\n        lexer.column = 1\n    lexer.position += len(lex)\n    lexer.set_error(f\'{lexer.lineno, lexer.column} -LexicographicError: EOF in comment\')\n```\n\nAnd to report general errors during the tokenization process we can use the `lexical_error` decorator.\n\n```python\n@g.lexical_error\ndef lexical_error(lexer):\n    lexer.set_error(lexer.line, lexer.column, f\'{lexer.lineno, lexer.column} -LexicographicError: ERROR "{lexer.token.lex}"\')\n    lexer.column += len(lexer.token.lex)\n    lexer.position += len(lexer.token.lex)\n```\n\n## Credits \n\nFor each recommendation or bug please write to alejandroklever.workon@gmail.com.\n',
    'author': 'Alejandro Klever',
    'author_email': 'alejandroklever4197@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alejandroklever/PyJapt',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
